;=========================================================================
; This MASM source implements a millisecond timer and a sleep procedure
; based on the timer.
;=========================================================================

.386

;----------------------------------------------------------------
; This macro implements a short I/O delay, used between accesses
; to the same I/O device to give the device time to respond.
; This delay may be necessary on very old systems; newer systems
; typically implement an appropriate delay in the chipset.
;----------------------------------------------------------------

IODELAY macro
  repeat 10
    jmp $+2
  endm
endm

;------------------------------------------------------------------
; The following data needs to be accessed from our system timer
; tick interrupt handler. In an interrupt handler the only segment
; register that always has a known value is CS, so to make the
; code work regardless of the memory model or startup code, we
; place the data in our code segment. Since we don't know where in
; the code segment it will end up, we include a jump to direct the
; execution path around it.
;------------------------------------------------------------------

.code

    jmp @F
    prevIsr8  dd 0
    tickCount dd 0
    counter   dw 0
  @@:

;=========================================================================

StartMsTimer proc uses ax bx dx es ds

    ;-----------------------------------------------------
    ; Hook the system timer tick interrupt (interrupt 8).
    ; The Get Interrupt Vector function returns the
    ; current vector in ES:BX.
    ;-----------------------------------------------------

    mov ax, 3508h
    int 21h
    mov WORD PTR prevIsr8, bx
    mov WORD PTR prevIsr8[2], es

    ;-------------------------------------------------------
    ; The Set Interrupt Vector function expects the address
    ; of the new handler in DS:DX. Since handler is in the
    ; code segment, DS must be temporarily loaded with the
    ; segment of the code segment.
    ;-------------------------------------------------------

    push cs
    pop ds
    mov dx, Isr8
    mov ax, 2508h
    int 21h

    ;----------------------------------------------------------
    ; Reprogram system timer 0 to provide a higher resolution.
    ; The system timers share a 1193182Hz clock, and system
    ; timer 0 is normally loaded with an initial count of zero.
    ; This causes it to count 65536 clock cycles between each
    ; interrupt, generating 1193182 / 65536 ~ 18.2 interrupts
    ; per second. An initial count of 1192 will produce an
    ; interrupt rate of ~1000.99 interrupts per second.
    ;
    ; Program system timer 0 for LSB then MSB, mode 3, binary.
    ; The system timer control word register is at I/O port
    ; 43h. The system timer control word is set as follows:
    ;   bit 7-6: 00     = timer 0
    ;   bit 5-4: 11     = R/W LSB then MSB
    ;   bit 3-1: 011    = mode 3, periodic square wave
    ;   bit 0:   0     	= binary
    ;----------------------------------------------------------

    cli
    mov al, 00110110b
    out 43h, al
    IODELAY
    mov ax, 1192
    out 40h, al       ; write LSB
    IODELAY
    mov al, ah
    out 40h, al       ; write MSB
    sti
    ret

StartMsTimer endp

;=========================================================================

StopMsTimer proc uses ax dx ds

    ;-----------------------------------
    ; Restore system timer 0 to normal.
    ;-----------------------------------

    cli
    mov al, 00110110b
    out 43h, al
    IODELAY
    xor al, al
    out 40h, al       ; write LSB
    IODELAY
    out 40h, al       ; write MSB
    sti

    ;-----------------------------------------
    ; Unhook the system timer tick interrupt.
    ;-----------------------------------------

    lds dx, prevIsr8
    mov ax, 2508h
    int 21h

    ret

StopMsTimer endp

;=========================================================================

;--------------------------------------------------------------
; This is the new handler for the system timer tick interrupt.
; The CS overrides are added here for clarity. MASM knows that
; the variables are defined in the code segment so it will add
; the necessary overrides here and elsewhere in this source.
;
; Note that a hardware interrupt handler must preserve all
; registers.
;--------------------------------------------------------------

Isr8 proc

    ;------------------------
    ; Update the tick count.
    ;------------------------

    inc cs:tickCount

    ;--------------------------------------------------
    ; Pass the interrupt to the previous interrupt 8
    ; handler every 55 interrupts. This will allow the
    ; system timer tick to function at close to its
    ; normal rate (actually ~.037% lower than normal).
    ;--------------------------------------------------

    inc cs:counter
    cmp cs:counter, 55
    jb  @F
    mov cs:counter, 0

    ;------------------------------------------------
    ; Chain to the previous handler with a far jump.
    ;------------------------------------------------

    jmp cs:prevIsr8

  @@:

    ;----------------------------------------------------------
    ; For a hardware interrupt the last handler in the handler
    ; chain must issue an EOI to the interrupt controller to
    ; prepare it for the next interrupt. The code above depends
    ; on the handler at the end of the chain to issue the EOI.
    ; Here, we act as the handler at the end of the chain,
    ; issue the EOI, and return from the interrupt.
    ;----------------------------------------------------------

    push ax
    mov al, 20h
    out 20h, al
    pop ax

    iret

Isr8 endp

;=========================================================================

;----------------------------------------------------------------
; This procedure delays for the specified number of milliseconds
; before returning. Note that this code does not allow for the
; tick count value rolling over. Since the tick count is a
; 32-bit value that starts at zero and increments ~1000.99 times
; per second, ~49.66 days from initialization it will roll over
; and start again at zero.
;----------------------------------------------------------------

Sleep proc uses eax dwMilliseconds:DWORD

    mov eax, tickCount
    add eax, dwMilliseconds
  @@:
    cmp tickCount, eax
    jb  @B
    ret

Sleep endp

;=========================================================================
