;=========================================================================
; This MASM source is essentially a simplified version of Milo Sedlaceks
; MULTIKEY keyboard handler, originally for QuickBASIC. It maintains an
; array of flags, indexed by key scan code, where each flag indicates the
; press/release-state of the key. It allows an application to respond to
; multiple keys simultaneously (thus the name), and avoids the delays
; inherent in the normal keyboard input functions.
;=========================================================================

;---------------------------------------------------
; We put the keyState array in the data segment for
; easy access from the program.
;---------------------------------------------------

.data

    keyState db 128 dup(0)

;=========================================================================

;---------------------------------------------------------------------
; The following data needs to be accessed from our Keyboard Intercept
; handler. In an interrupt handler the only segment register that
; always has a known value is CS, so to make the code work regardless
; of the memory model or startup code, we place the data in our code
; segment. Since we don't know where in the code segment it will end
; up, we include a jump to direct the execution path around it.
;---------------------------------------------------------------------

.code

    jmp @F
    prevIsr15 dd 0
    ourDS     dw 0
  @@:

;=========================================================================

;-----------------------------------------------------------------------
; This is our Keyboard Intercept (Interrupt 15h, function 4Fh) handler.
; The BIOS Interrupt 9 handler calls this handler each time a key is
; pressed or released, with the Keyboard Intercept function number in
; AH and the key scan code in AL.
;
; The CS overrides are added here for clarity. MASM knows that the
; variables are defined in the code segment so it will add the necessary
; overrides here and elsewhere in this source.
;
; Note that a hardware interrupt handler (or a handler called by a
; hardware interrupt handler) must preserve all registers.
;-----------------------------------------------------------------------

Isr15 proc

    ;--------------------------------------------------
    ; Chain to the previous handler if the function is
    ; anything other than the Keyboard Intercept.
    ;--------------------------------------------------

    cmp ah, 4fh
    je  @F
    jmp cs:prevIsr15

  @@:

    ;--------------------------------------------------
    ; Preserve the registers that we need to alter and
    ; set DS to the program's data segment.
    ;--------------------------------------------------

    push ax
    push bx
    push dx
    push ds
    push cs:ourDS
    pop ds

    ;-----------------------------------------------------------------
    ; Read the scan code directly from the keyboard controller's data
    ; port and update the keyState array. Bit 7 of the value is set
    ; for a break code (sent when a key is released), and cleared for
    ; a make code (sent when a key is pressed). The value stored in
    ; the array is 1 for a make code and 0 for break code.
    ;
    ; Note that bit 7 must be cleared before the value can be used to
    ; index the keyState array.
    ;-----------------------------------------------------------------

    in  al, 60h
    xor dl, dl
    test al, 80h
    jnz @F
    mov dl, 1
  @@:
    and al, 7fh
    xor bx, bx
    mov bl, al
    mov byte ptr keyState[bx], dl

    ;------------------------------------------------------------
    ; A keyboard intercept handler controls what the Interrupt 9
    ; handler does with the scan code by setting or clearing the
    ; Interrupt 9 handlers carry flag. Because the flags are
    ; preserved across the interrupt call/return sequence, with
    ; the INT instruction pushing the flags onto the stack and
    ; the IRET instruction popping them from the stack, to modify
    ; the Interrupt 9 handlers carry flag we need to modify the
    ; copy of the flags on the stack. Here we need to clear the
    ; carry flag so the Interrupt 9 handler will know to ignore
    ; the scan code, instead of processing the keystroke and
    ; placing it in the key buffer.
    ;
    ; The contents of the stack at this point, relative to
    ; SP, are:
    ;   [SP+12] flags preserved by INT instruction
    ;   [SP+10] CS preserved by INT instruction
    ;   [SP+8]  IP preserved by INT instruction
    ;   [SP+6]  preserved AX
    ;   [SP+4]  preserved BX
    ;   [SP+2]  preserved DX
    ;   [SP+0]  preserved DS
    ;
    ; The carry flag is bit 0 of the flags register.
    ;------------------------------------------------------------

    mov bx, sp
    mov ax, ss:[bx+12]
    and ax, 0fffeh
    mov ss:[bx+12], ax

    ;----------------------------------------------------------------
    ; Recover the preserved registers and return from the interrupt.
    ;----------------------------------------------------------------

    pop ds
    pop dx
    pop bx
    pop ax

    iret

Isr15 endp

;=========================================================================

StartMultikey proc uses ax bx dx es ds

    ;------------------------------------------------
    ; Initialize ourDS so the handler can use it to
    ; access the keyState array in the data segment.
    ;------------------------------------------------

    push ds
    pop ourDS

    ;----------------------------------------------
    ; Hook the Keyboard Intercept interrupt (15h).
    ; The Get Interrupt Vector function returns the
    ; current vector in ES:BX.
    ;----------------------------------------------

    mov ax, 3515h
    int 21h
    mov WORD PTR prevIsr15, bx
    mov WORD PTR prevIsr15[2],es


    ;-------------------------------------------------------
    ; The Set Interrupt Vector function expects the address
    ; of the new handler in DS:DX. Since handler is in the
    ; code segment, DS must be temporarily loaded with the
    ; segment of the code segment.
    ;-------------------------------------------------------

    push cs
    pop ds
    mov ax, 2515h
    mov dx, Isr15
    int 21h

    ret

StartMultikey endp

;=========================================================================

StopMultikey proc uses ax dx ds

    ;------------------------------------------
    ; Unhook the Keyboard Intercept interrupt.
    ;------------------------------------------

    push ds
    lds dx, prevIsr15
    mov ax,2515h
    int 21h
    pop ds
    ret

StopMultikey endp

;=========================================================================

